/**/

#include <obstack.h>
#include <glpk.h>

#include "rscommon.h"

DEBUG_DEFINE_DEBUGF(value);

typedef struct { int mass, volu; } CommodInfo;
static int commodstablesz;
static CommodInfo *commodstable;

static sqlite3_stmt *ss_ipair_dist, *ss_ipair_trades;

#define MAX_LEGS (MAX_ROUTELEN-1)

typedef struct {
  int commodid, src_price, src_qty, dst_price, dst_qty;
} Trade;

#define TRADES_PER_BLOCK 10

typedef struct TradesBlock{
  struct TradesBlock *next;
  Trade t[TRADES_PER_BLOCK];
} TradesBlock;

typedef struct {
  double distance_loss_factor;
  int ntrades;
  TradesBlock *trades;
} IslandPair;

int nislands;
IslandPair ***ipairs; /* ipairs[sislandid][dislandid] */

typedef struct IslandTradeEnd {
  struct IslandTradeEnd *next;
  int commodid, price;
  int qty, rownum;
} IslandTradeEnd, *IslandDirnTradeEnds;

typedef struct {
  int islandid;
  IslandDirnTradeEnds collect, deliver;
} IslandTradeEnds;

static struct obstack ites_obstack;
static LPX *lp;

static int nconstraint_rows;
static int constraint_rows[1+2+3*MAX_LEGS];
static double constraint_coeffs[1+2+3*MAX_LEGS];
      /* dummy0, src, dst, for_each_leg( [mass], [volume], [capital] ) */

static void add_constraint(int row, double coefficient) {
  nconstraint_rows++; /* glpk indices start from 1 !!! */
  constraint_rows  [nconstraint_rows]= row;
  constraint_coeffs[nconstraint_rows]= coefficient;
}

static void avail_constraint(const Trade *t, IslandDirnTradeEnds *trades,
			     int price, int qty, const char *srcdst) {
  /* find row number of trade availability constraint */
  IslandTradeEnd *search;

  for (search= *trades; search; search=search->next)
    if (search->commodid==t->commodid && search->price==price) {
      assert(search->qty == qty);
      goto found;
    }
  /* not found, add new row */
  search= obstack_alloc(&ites_obstack, sizeof(*search));
  search->commodid= t->commodid;
  search->price= price;
  search->qty= qty;
  search->rownum= lpx_add_rows(lp, 1);
  search->next= *trades;
  if (DEBUGP(value)) {
    char *name= masprintf("%s_commod%d_price%d",srcdst,t->commodid,price);
    lpx_set_row_name(lp,search->rownum,name);
    free(name);
  }
  *trades= search;
 found:;
  int rownum= search->rownum;

  lpx_set_row_bnds(lp, rownum, LPX_UP, 0, qty);
  add_constraint(rownum, 1.0);
}

static int setup_leg_constraints(double max_thing, int legs, const char *wh) {
  int leg, startrow;
  if (max_thing < 0 || !legs) return -1;
  startrow= lpx_add_rows(lp, legs);
  for (leg=0; leg<nislands-1; leg++) {
    int row= leg+startrow;
    lpx_set_row_bnds(lp, row, LPX_UP, 0, max_thing);
    if (DEBUGP(value)) {
      char *name= masprintf("max_leg%d_%s",leg,wh);
      lpx_set_row_name(lp,row,name);
      free(name);
    }
  }
  return startrow;
}

static void add_leg_c(int startrow, int leg, double value) {
  if (startrow<=0) return;
  assert(value > 0);
  add_constraint(startrow+leg, value);
}

static IslandPair *ipair_get(int si, int di) {
  IslandPair *ip, **ipa;

  assert(si < nislands);
  assert(di < nislands);

  if (!(ipa= ipairs[si])) {
    ipa= ipairs[si]= mcalloc(sizeof(*ipa) * nislands);
  }
  if ((ip= ipa[di]))
    return ip;

  ipa[di]= ip= mmalloc(sizeof(*ip));
  ip->ntrades= 0;
  ip->trades= 0;
  int inblock= TRADES_PER_BLOCK;
  TradesBlock *block= 0;

  debugf("VALUE ipair_get(%d,%d) running...\n", si,di);
  SQL_MUST( sqlite3_bind_int(ss_ipair_dist, 1, si) );
  SQL_MUST( sqlite3_bind_int(ss_ipair_dist, 2, di) );
  assert(SQL_STEP(ss_ipair_dist));
  int dist= sqlite3_column_int(ss_ipair_dist, 0);
  ip->distance_loss_factor= pow(distance_loss_factor_per_league, dist);
  sqlite3_reset(ss_ipair_dist);

  SQL_MUST( sqlite3_bind_int(ss_ipair_trades, 1, si) );
  SQL_MUST( sqlite3_bind_int(ss_ipair_trades, 2, di) );

  while (SQL_STEP(ss_ipair_trades)) {
    if (inblock == TRADES_PER_BLOCK) {
      block= mmalloc(sizeof(*block));
      block->next= ip->trades;
      ip->trades= block;
      inblock= 0;
    }
    int *irp, i;
    for (i=0, irp=&block->t[inblock].commodid; i<5; i++, irp++)
      *irp= sqlite3_column_int(ss_ipair_trades, i);
    ip->ntrades++;
    inblock++;
  }
  if (inblock < TRADES_PER_BLOCK)
    block->t[inblock].commodid= -1;

  sqlite3_reset(ss_ipair_trades);
  
  return ip;
}

void value_route(int nislands, const int *islands) {
  int s,d;

  /* We need to construct the LP problem.  GLPK talks
   * about rows and columns, which are numbered from 1.
   *
   * Each column is a `structural variable' ie one of the entries in
   * the objective function.  In our case the set of structural
   * variable is, for each port, the set of Trades which collect at
   * that island.  (We use `port' to mean `specific visit to an
   * island' so if an island appears more than once so do its trades.)
   * We don't need to worry about crossing with all the possible
   * delivery locations as we always deliver on the first port.
   * We will call such a structural variable a Flow, for brevity.
   *
   * We iterate over the possible Flows adding them as columns as we
   * go, and also adding their entries to the various constraints.
   *
   * Each row is an `auxiliary variable' ie one of the constraints.
   * We have two kinds of constraint:
   *   - mass/volume/capital: one constraint for each sailed leg
   *       (unless relevant constraint is not satisfied)
   *   - quantity of commodity available for collection
   *       or delivery at particular price and island
   * The former are numbered predictably: we have first all the mass
   * limits, then all the volume limits, then all the capital limits
   * (as applicable) - one for each leg, ie one for each entry
   * in islands except the first.
   *
   * The latter are added as needed and the row numbers are stored in
   * a data structure for later reuse.
   */

  assert(nislands >= 1);

  char *free_to_reset_obstack= obstack_alloc(&ites_obstack, 1);

  int nites=0;
  IslandTradeEnds ites[nislands], *iteps[nislands];

  for (s=0; s<nislands; s++) {
    IslandTradeEnds *ite;
    int si= islands[s];
    int i;
    for (i=0, ite=ites; i<nites; i++, ite++)
      if (ite->islandid==si)
	goto found;
    /* not found, add new */
    assert(ite == &ites[nites]);
    ite->islandid= si;
    ite->collect= ite->deliver= 0;
    nites++;
  found:
    iteps[s]= ite;
  }

  assert(!lp);
  lp= lpx_create_prob();
  lpx_set_obj_dir(lp, LPX_MAX);
  if (DEBUGP(value)) {
    lpx_set_prob_name(lp,(char*)"value_route");
    lpx_set_obj_name(lp,(char*)"profit");
  }

  int legs= nislands-1;
  int mass_constraints= setup_leg_constraints(max_mass, legs, "mass");
  int volu_constraints= setup_leg_constraints(max_volu, legs, "volu");
  int capi_constraints= setup_leg_constraints(max_capi, legs, "capi");

  double delay_slot_loss_factor= 1.0;
  for (s=0;
       s<nislands;
       s++, delay_slot_loss_factor *= LOSS_FACTOR_PER_DELAY_SLOT) {
    int si= islands[s];
    
    for (d=s; d<nislands; d++) {
      int di= islands[d];
      int already_d;
      for (already_d=s+1; already_d<d; already_d++)
	if (islands[already_d] == di)
	  /* visited this island already since we left s, uninteresting */
	  goto next_d;

      if (d>s && di==si)
	/* route has returned to si, no need to think more about s */
	goto next_s;

      /*----- actually add these trades to the LP problem -----*/
      
      IslandPair *ip= ipair_get(islands[s], islands[d]);
      TradesBlock *block= ip->trades;
      int tradestodo= ip->ntrades;
      if (!tradestodo)
	goto next_d;

      int inblock= 0;
      int col= lpx_add_cols(lp,ip->ntrades);

      double loss_factor= delay_slot_loss_factor * ip->distance_loss_factor;

      while (tradestodo-- >0) {
	if (inblock >= TRADES_PER_BLOCK) {
	  block= block->next;
	  inblock= 0;
	}
	Trade *t= &block->t[inblock++];

	debugf("  TRADE %d#%d..%d#%d %d %d-%d\n",
	       si,s, di,d, t->commodid, t->src_price, t->dst_price);

	nconstraint_rows=0;

	avail_constraint(t, &iteps[s]->collect, t->src_price,t->src_qty,"src");
	avail_constraint(t, &iteps[d]->deliver, t->dst_price,t->dst_qty,"dst");

	int leg;
	for (leg=s; leg<d; leg++) {
	  add_leg_c(mass_constraints, leg, commodstable[t->commodid].mass);
	  add_leg_c(volu_constraints, leg, commodstable[t->commodid].volu);
	  add_leg_c(capi_constraints, leg, t->src_price);
	}

	lpx_set_col_bnds(lp, col, LPX_LO, 0, 0);
	lpx_set_obj_coef(lp, col,
			 (t->dst_price - t->src_price) * loss_factor);
	lpx_set_mat_col(lp, col, nconstraint_rows,
			constraint_rows, constraint_coeffs);
	if (DEBUGP(value)) {
	  char *name= masprintf("trade_commod%d_port%d_at%d_port%d_at%d",
				t->commodid, s, t->src_price, d, t->dst_price);
	  lpx_set_col_name(lp, col, name);
	  free(name);
	}
      } /* while (tradestodo-- >0) */
      
      /*----- that's done adding these trades to the LP problem -----*/
      
    next_d:;
    } /* for (d) */
  next_s:;
  } /* for (s) */

  lpx_delete_prob(lp);
  lp= 0;

  obstack_free(&ites_obstack, free_to_reset_obstack);
}

void setup_value(void) {
  sqlite3_stmt *sst;
  int i;

#define obstack_chunk_alloc mmalloc
#define obstack_chunk_free free
  obstack_init(&ites_obstack);

  nislands= sql_single_int("SELECT max(islandid) FROM islands") + 1;
  debugf("VALUE nislands=%d\n",nislands);

  commodstablesz= sql_single_int("SELECT max(commodid) FROM commods") + 1;
  commodstable= mmalloc(sizeof(*commodstable)*commodstablesz);
  for (i=0; i<commodstablesz; i++)
    commodstable[i].mass= commodstable[i].volu= -1;

  SQL_MUST( sqlite3_prepare(db,
     "SELECT commodid,unitmass,unitvolume FROM commods", -1,&sst,0) );
  while (SQL_STEP(sst)) {
    int id= sqlite3_column_int(sst,0);
    assert(id>=0 && id<commodstablesz);
    commodstable[id].mass= sqlite3_column_int(sst,1);
    commodstable[id].volu= sqlite3_column_int(sst,2);
  }
  sqlite3_finalize(sst);

  SQL_MUST( sqlite3_prepare(db,
     "SELECT\n"
     " sell.commodid		commodid,\n"
     " sell.price		src_price,\n"
     " sum(sell.qty)		src_qty,\n"
     " buy.price		dst_price,\n"
     " sum(buy.qty) 		dst_qty\n"
     " FROM sell JOIN buy\n"
     "   ON sell.commodid = buy.commodid\n"
     "  AND buy.price > sell.price\n"
     " WHERE sell.islandid=?\n"
     "   AND buy.islandid=?\n"
     " GROUP BY sell.commodid, sell.price, buy.price",
			    -1, &ss_ipair_trades, 0) );

  SQL_MUST( sqlite3_prepare(db,
     " SELECT dist from dists where aiid=? and biid=?",
			    -1, &ss_ipair_dist, 0) );

  ipairs= mcalloc(sizeof(*ipairs) * nislands);
}
